# vi: ts=4 expandtab
#
# Copyright (C) 2014 Canonical Ltd.
#
# Author: Blake Rouse <blake.rouse@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

import os
import shutil
import yaml

from builder.osystems import OperatingSystemRegistry
from builder.utils import (
    ProcessExecutionError,
    subp,
    )


def install_dependencies():
    """Installs all of the dependencies that the script requires."""
    subp([
        'apt-get', 'install', '-y',
        'kpartx',
        'qemu-utils',
        'kvm',
        'libvirt-bin',
        'virtinst',
        'genisoimage',
        'dos2unix',
        'unzip',
        ])


def is_installed(cmd):
    """Checks if the command is installed."""
    try:
        subp(['which', cmd], capture=True)
    except ProcessExecutionError:
        return False
    return True


def has_all_commands():
    """Return True if all commands are available."""
    commands = [
        'kpartx',
        'qemu-img',
        'virt-install',
        'virsh',
        'dos2unix',
        'kvm-spice',
        'genisoimage',
        'unzip',
        ]
    for cmd in commands:
        if not is_installed(cmd):
            return False
    return True


def main(args):
    """Main execution of the application.

    :param args: parsed command line arguments
    :returns: application exit code
    """
    # Check that have root privledges
    if os.geteuid() != 0:
        print('Error: must run with root privileges.')
        return 1

    # Check for install deps
    if args.install_deps:
        install_dependencies()
        return 0
    if not has_all_commands():
        print('Error: missing required packages (run with --install-deps).')
        return 1

    # Get the osystem to build
    osystem = OperatingSystemRegistry[args.osystem]

    # Check osystem supports architecture
    arch = args.arch
    if arch not in osystem.arches:
        print(
            "Error: %s does't support %s architecture." % (osystem.name, arch))
        return 1

    # Build
    osystem.build_image(args)
    return 0
