# vi: ts=4 expandtab
#
# Copyright (C) 2014 Canonical Ltd.
#
# Author: Blake Rouse <blake.rouse@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

import os
import shutil

from builder.osystems import VirtInstallOS


class CentOS(VirtInstallOS):
    """Builds the CentOS image for both amd64 and i386. Uses virt-install
    to perform this installation process."""

    name = "centos"
    arches = ["i386", "amd64"]
    os_type = "linux"
    os_variant = "rhel5"
    disk_size = 5
    nic_model = "virtio"
    install_location = ""

    def full_name(self, params):
        return 'centos%s-%s' % (params.centos_edition, params.arch)

    def build_parser(self, parser):
        """Add RHEL specific arguments."""
        parser.add_argument(
            '--centos-edition', default='6',
            help="CentOS edition to generate. (Default: 6)")

    def validate_params(self, params):
        """Validates the command line parameters."""
        if params.centos_edition not in ['6', '7']:
            raise OperatingSystemError(
                "Unknown CentOS edition: %s." % params.centos_edition)
        if params.centos_edition == '7' and params.arch == 'i386':
            raise OperatingSystemError(
                "Cannot generate CentOS 7 for i386, as only amd64 is "
                "supported.")

    def modify_mount(self, mount_path):
        """Install the curtin directory into mount point."""
        path = None
        if self.edition == '6':
            path = self.get_contrib_path('centos6/curtin')
        elif self.edition == '7':
            path = self.get_contrib_path('centos7/curtin')
        if not os.path.exists(path):
            return
        opt_path = os.path.join(mount_path, 'curtin')
        shutil.copytree(path, opt_path)

    def build_image(self, params):
        self.validate_params(params)
        self.edition = params.centos_edition
        if self.edition == '6':
            arch = params.arch
            if arch == 'i386':
                self.install_location = (
                    "http://mirrors.kernel.org/centos/6/os/i386")
                self.extra_arguments = (
                    "console=ttyS0 ks=file:/centos6-i386.ks text utf8")
                self.initrd_inject = self.get_contrib_path(
                    "centos6/centos6-i386.ks")
            elif arch == 'amd64':
                self.install_location = (
                    "http://mirrors.kernel.org/centos/6/os/x86_64")
                self.extra_arguments = (
                    "console=ttyS0 ks=file:/centos6-amd64.ks text utf8")
                self.initrd_inject = self.get_contrib_path(
                    "centos6/centos6-amd64.ks")
        elif self.edition == '7':
            self.install_location = (
                "http://mirrors.kernel.org/centos/7/os/x86_64")
            self.extra_arguments = (
                "console=ttyS0 inst.ks=file:/centos7-amd64.ks text "
                "inst.cmdline inst.headless")
            self.initrd_inject = self.get_contrib_path(
                "centos7/centos7-amd64.ks")
        super(CentOS, self).build_image(params)
