# vi: ts=4 expandtab
#
# Copyright (C) 2014 Canonical Ltd.
#
# Author: Blake Rouse <blake.rouse@canonical.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import (
    absolute_import,
    print_function,
    unicode_literals,
    )

import os
import re
import glob
import shutil

from builder import (
    net,
    utils,
    )
from builder.osystems import (
    OperatingSystem,
    OperatingSystemError,
    )
from tempita import Template


EDITIONS = {
    'win2008r2': "Windows Server 2008 R2 SERVERSTANDARD",
    'win2008hvr2': "Windows Server 2008 R2 SERVERHYPERCORE",
    'win2012': "Windows Server 2012 SERVERSTANDARD",
    'win2012hv': "Hyper-V Server 2012 SERVERHYPERCORE",
    'win2012r2': "Windows Server 2012 R2 SERVERSTANDARD",
    'win2012hvr2': "Hyper-V Server 2012 R2 SERVERHYPERCORE",
    }


class WindowsOS(OperatingSystem):
    """Builds the Windows image."""

    name = "windows"
    arches = ["i386", "amd64"]

    def build_parser(self, parser):
        """Add Windows specific arguments."""
        parser.add_argument(
            '--windows-iso',
            help="Path to Windows installation ISO.")
        parser.add_argument(
            '--windows-edition',
            help="Windows edition to install from the ISO.")
        parser.add_argument(
            '--windows-license-key',
            help="Windows license key to embed into generated image.")
        parser.add_argument(
            '--windows-updates', action='store_true',
            help=(
                "Install all Windows updates into generated image. "
                "(Requires access to microsoft.com)"))
        parser.add_argument(
            '--windows-drivers',
            help=(
                "Folder containing drivers to be injected into the Windows "
                "installation before the image is generated."))
        parser.add_argument(
            '--windows-language',
            default='en-US',
            help="Windows installation language. Default: en-US")

    def validate_params(self, params):
        """Validates the command line parameters."""
        iso = params.windows_iso
        if iso is None:
            raise OperatingSystemError(
                "Windows requires the --windows-iso option.")
        if not os.path.exists(iso):
            raise OperatingSystemError(
                "Failed to access Windows ISO at: %s" % iso)
        edition = params.windows_edition
        if edition is None or edition == '':
            raise OperatingSystemError(
                "Windows requires the --windows-edition option.")
        if edition not in EDITIONS.keys():
            raise OperatingSystemError(
                "Invalid Windows edition, should be one of %s." % (
                    EDITIONS.keys()))
        license_key = params.windows_license_key
        if license_key is not None and license_key != '':
            if not self.validate_license_key(license_key):
                raise OperatingSystemError(
                    "Invalid Windows license key.")
        drivers = params.windows_drivers
        if drivers is not None and not os.path.isdir(drivers):
            raise OperatingSystemError(
                "Invalid driver path: %s" % drivers)

    def validate_license_key(self, license_key):
        """Validates that license key is in the correct format. It does not
        validate, if that license key will work with the selected edition of
        Windows."""
        r = re.compile('^([A-Za-z0-9]{5}-){4}[A-Za-z0-9]{5}$')
        return r.match(license_key)

    def load_unattended_template(self):
        """Loads the unattended template that is used for installation."""
        path = self.get_contrib_path('Autounattend.xml')
        return Template.from_filename(path)

    def write_unattended(self, output_path, arch, edition, language,
                         license_key=None, enable_updates=False):
        """Outputs the effective unattended.xml file that will be used by
        Windows during the installation."""
        template = self.load_unattended_template()
        image_name = EDITIONS[edition]
        # Windows doesn't accept i386, instead that maps to x86.
        if arch == 'i386':
            arch = 'x86'
        output = template.substitute(
            arch=arch, image_name=image_name, language=language,
            license_key=license_key, enable_updates=enable_updates)
        with open(output_path, 'w') as stream:
            for line in output.splitlines():
                stream.write("%s\r\n" % line)

    def create_floppy_disk(self, output_path):
        """Creates an empty floppy disk, formatted with vfat."""
        utils.subp([
            'dd', 'if=/dev/zero',
            'of=%s' % output_path,
            'bs=1024', 'count=1440',
            ])
        utils.subp([
            'mkfs.vfat',
            output_path
            ])

    def prepare_floppy_disk(self, workdir, arch, edition, language,
                            license_key=None, enable_updates=False):
        """Prepares the working directory with Autounattend.vfd."""
        # Create the disk
        vfd_path = os.path.join(workdir, 'Autounattend.vfd')
        self.create_floppy_disk(vfd_path)

        # Mount the disk
        mount_path = os.path.join(workdir, 'vfd_mount')
        os.mkdir(mount_path)
        utils.subp([
            'mount',
            '-t', 'vfat',
            '-o', 'loop',
            vfd_path, mount_path,
            ])

        # Place the generated Autounattend.xml file
        xml_path = os.path.join(mount_path, 'Autounattend.xml')
        try:
            self.write_unattended(
                xml_path, arch, edition, language,
                license_key=license_key, enable_updates=enable_updates)
        finally:
            utils.subp(['umount', mount_path])
            os.rmdir(mount_path)
        return vfd_path

    def download_cloudbase_init(self, workdir, arch):
        """Downloads cloudbase init."""
        output_path = os.path.join(workdir, 'cloudbase_init.msi')
        if arch == 'amd64':
            msi_file = "CloudbaseInitSetup_Beta_x64.msi"
        elif arch == 'i386':
            msi_file = "CloudbaseInitSetup_Beta_x86.msi"
        download_path = "http://www.cloudbase.it/downloads/" + msi_file

        # Remove me, testing only
        tmp_path = os.path.join('/tmp', msi_file)
        if os.path.exists(tmp_path):
            shutil.copyfile(tmp_path, output_path)
            return output_path

        utils.subp([
            'wget',
            '-O', output_path,
            download_path
            ])
        return output_path

    def download_ps_windows_update(self, workdir):
        """Downloads the PSWindowsUpdate package."""
        output_path = os.path.join(workdir, 'pswindowsupdate.zip')
        download_path = (
            "http://gallery.technet.microsoft.com/scriptcenter/"
            "2d191bcd-3308-4edd-9de2-88dff796b0bc/file/41459/43/"
            "PSWindowsUpdate.zip")
        utils.subp([
            'wget',
            '-O', output_path,
            download_path
            ])
        return output_path

    def unzip_archive(self, src, dest):
        """Un-zips an archive into destination."""
        utils.subp([
            'unzip', '-q',
            src,
            '-d', dest,
            ])

    def create_iso(self, output, source):
        """Creates iso at output, containing files at source."""
        utils.subp([
            'genisoimage',
            '-o', output,
            '-V', 'SCRIPTS',
            '-J', source
            ])

    def build_install_iso(self, workdir, arch, with_updates=False,
                          drivers_path=None):
        """Builds the iso that is mounted to Windows, to complete the
        installation process."""
        install_path = os.path.join(workdir, 'install')
        os.mkdir(install_path)

        # Download cloudbase-init into install/cloudbase
        cloudbase_dir = os.path.join(install_path, 'cloudbase')
        os.mkdir(cloudbase_dir)
        self.download_cloudbase_init(cloudbase_dir, arch)

        # Copy contrib scripts into install/scripts
        contrib_path = self.get_contrib_path('scripts')
        scripts_path = os.path.join(install_path, 'scripts')
        shutil.copytree(contrib_path, scripts_path)

        # Copy the drivers if provided
        if drivers_path is not None:
            shutil.copytree(
                drivers_path,
                os.path.join(install_path, 'infs'))

        # Place PSWindowsUpdate modules if using with_updates
        if with_updates:
            zip_path = self.download_ps_windows_update(workdir)
            self.unzip_archive(zip_path, install_path)

        # Create the iso
        output_iso = os.path.join(workdir, 'install.iso')
        self.create_iso(output_iso, install_path)
        shutil.rmtree(install_path)
        return output_iso

    def create_disk_image(self, output_path, size):
        """Creates the disk image that Windows will install to."""
        utils.subp([
            'qemu-img', 'create',
            '-f', 'raw',
            output_path, size
            ])

    def spawn_vm(self, ram, vcpus, cdrom, floppy, install_iso, disk,
                 tap=None):
        """Spawns the qemu vm for Windows to install."""
        args = [
            'kvm-spice',
            '-m', '%s' % ram, '-smp', vcpus,
            '-cdrom', cdrom,
            '-drive', 'file=%s,index=3,media=cdrom' % install_iso,
            '-fda', floppy,
            ]
        if tap is not None:
            mac = net.get_random_qemu_mac()
            args.extend([
                '-device', 'rtl8139,netdev=net00,mac=%s' % mac,
                '-netdev', 'type=tap,id=net00,script=no,downscript=no,ifname=%s' % tap,
                ])
        args.extend([
            '-boot', 'd', '-vga', 'std',
            '-k', 'en-us',
            # Debug *Remove*
            '-vnc', ':1',
            disk
            ])
        utils.subp(args)

    def mount_partition(self, workdir, disk_path, partition):
        """Mounts the parition from the disk."""
        mount_path = os.path.join(workdir, 'disk_mount')
        os.mkdir(mount_path)
        utils.mount_loop(disk_path, mount_path, partition)
        return mount_path

    def umount_partition(self, disk_path, target, partition):
        """Un-mounts the target, marks ntfs as clean, and removes loopback."""
        utils.subp(['umount', target])
        devs = utils.kpartx_list(disk_path)
        utils.subp(['ntfsfix', '-d', devs[partition]])
        utils.kpartx_del(disk_path)
        os.rmdir(target)

    def convert_to_unix(self, file_path):
        """Converts file to unix to easily view."""
        try:
            utils.subp(['dos2unix', file_path])
        except utils.ProcessExecutionError:
            pass

    def check_success(self, mount_path, save_error_path):
        """Checks for success.touch and for error_log.txt."""
        error_log_path = os.path.join(mount_path, 'error_log.txt')
        if os.path.exists(error_log_path):
            shutil.copy(error_log_path, save_error_path)
            self.convert_to_unix(save_error_path)
            raise OperatingSystemError((
                'Windows installation failed, '
                'output placed %s.') % save_error_path)
        success_path = os.path.join(mount_path, 'success.tch')
        if not os.path.exists(success_path):
            raise OperatingSystemError(
                'Windows installation failed with an unknown reason.')
        os.unlink(success_path)

    def install_curtin(self, mount_path):
        """Installs the curtin scripts folder into the root."""
        src_path = self.get_contrib_path('curtin')
        curtin_path = os.path.join(mount_path, 'curtin')
        shutil.copytree(src_path, curtin_path)

    def remove_serial_log_from_cloudbase_init(self, target):
        """Removes serial log output from the configuration files for
        cloudbase-init."""
        cloudbase_init_cfg = os.path.join(
            target,
            "Program Files (x86)",
            "Cloudbase Solutions",
            "Cloudbase-Init",
            "conf",
            "cloudbase-init.conf")
        cloudbase_init_unattended_cfg = os.path.join(
            target,
            "Program Files (x86)",
            "Cloudbase Solutions",
            "Cloudbase-Init",
            "conf",
            "cloudbase-init-unattend.conf")

        configs = [cloudbase_init_cfg, cloudbase_init_unattended_cfg]
        for config in configs:
            with open(config, 'r') as stream:
                data = stream.read()
                data = data.replace(
                    'logging_serial_port_settings=COM1,115200,N,8\r\n', '')
            with open(config, 'w') as stream:
                stream.write(data)

    def qemu_convert(self, disk_path, output_path):
        """Converts the disk path to output path."""
        utils.subp([
            'qemu-img', 'convert',
            '-O', 'raw',
            disk_path, output_path,
            ])

    def create_tarball(self, disk_path, output_path):
        """Creates tarball of the disk."""
        disk_dir = os.path.dirname(os.path.abspath(disk_path))
        disk_filename = os.path.basename(disk_path)
        utils.subp([
            'tar', 'Szcf',
            output_path,
            '-C', disk_dir,
            disk_filename,
            ])

    def build_image(self, params):
        self.validate_params(params)

        output = params.output
        interface = params.interface
        ram = params.ram
        vcpus = params.vcpus
        arch = params.arch
        iso = params.windows_iso
        edition = params.windows_edition
        license_key = params.windows_license_key
        updates = params.windows_updates
        drivers_path = params.windows_drivers
        language = params.windows_language

        # Create work space
        with utils.tempdir() as workdir:

            # Build the install.iso
            install_iso = self.build_install_iso(
                workdir, arch, with_updates=updates, drivers_path=drivers_path)

            # Create the floppy with the Autounattend.xml
            floppy_path = self.prepare_floppy_disk(
                workdir, arch, edition, language, license_key=license_key,
                enable_updates=updates)

            # Create the disk image
            disk_path = os.path.join(workdir, 'output.img')
            self.create_disk_image(disk_path, '16G')

            # Create tap device, if installing Windows updates
            # as the VM needs access to microsoft.com
            tap_name = None
            if updates:
                tap_name = net.create_tap(interface)

            try:
                # Start the Windows installation
                self.spawn_vm(
                    ram, vcpus, iso, floppy_path, install_iso, disk_path,
                    tap=tap_name)
            finally:
                # Destroy the tap
                if tap_name is not None:
                    net.delete_tap(tap_name)

            # Installation has finished, mount the disk
            mount_path = self.mount_partition(workdir, disk_path, 1)

            try:
                # Create the output directory
                if not os.path.exists(output):
                    os.mkdir(output)

                # Check that installation went as expected
                error_filename = 'windows-%s-%s-error.log' % (edition, arch)
                save_error_path = os.path.join(output, error_filename)
                self.check_success(mount_path, save_error_path)

                # Install the curtin scripts into the root
                self.install_curtin(mount_path)

                # Remove serial output from cloudbase-init.conf
                self.remove_serial_log_from_cloudbase_init(mount_path)

            finally:
                # Unmount and clean
                self.umount_partition(disk_path, mount_path, 1)

            # Convert to raw, to save on some sparse space
            clean_disk_path = os.path.join(workdir, 'clean-output.img')
            self.qemu_convert(disk_path, clean_disk_path)
            os.unlink(disk_path)

            # Create the tarball of raw image
            tarball_path = os.path.join(workdir, 'output.ddtgz')
            self.create_tarball(clean_disk_path, tarball_path)

            # Place into output
            output_filename = 'windows-%s-%s-root-dd' % (edition, arch)
            output_path = os.path.join(output, output_filename)
            shutil.move(tarball_path, output_path)
